/* topxloc.p
 *
 * process a profiler output file and produce a report of the top X lines of code in terms of execution time consumed
 *
 *    pro -p lib/topxloc.p -param tmp/something.prf
 *
 *    pro -p lib/topxloc.p -param tmp/something.prf|rpt/something.rpt|topX|debugLevel
 *
 * this is a completely stand-alone report, it depends on nothing other than an arbitrarily named profiler output file
 *
 * bug in oe12.2.3 results in no profiling data, must upgrade to 12.2.4 or later
 *  https://knowledgebase.progress.com/articles/Knowledge/profiler-system-handle-write-data-does-not-write-any-output/
 *
 */

define variable inputFileName   as character no-undo label "Input".
define variable outputFileName  as character no-undo label "Output".
define variable topLines        as integer   no-undo.
define variable debugLevel      as integer   no-undo.		/* set > 0 to output module list and timing summary	*/

define variable dateFormat      as character no-undo.		/* remember the local value before changing it		*/
define variable numberSep       as character no-undo.		/* so that we can restore it when we are done		*/

define variable keepReading     as logical no-undo.

define stream inStrm.

define temp-table tt_header					/* == header section ==					*/
  field profileDate as date format "99/99/99"
  field description as character format "x(60)" label "Description"
  field profileTime as character 				/* session time - hh:mm:ss at output 			*/
  field unused      as character				/* unused						*/
  field profileJSON as character				/* JSON data (ignored for now)				*/
  index idx-profileDate is primary profileDate
.

define temp-table tt_module					/* == module section ==					*/
  field moduleId     as integer format ">>>>>9"			/* module id						*/
  field moduleName   as character format "x(40)"		/* module name - "ipName sourceName"			*/
  field debugName    as character format "x(40)"		/* debug listing name					*/
  field rcodeCRC     as integer					/* r-code CRC value (v2?)				*/
  field procedureDef as integer					/* line# of the IP or method definition	(v3)		*/
  field procedureSig as character				/* call signature (v3)					*/
  /***/
  field srcName     as character format "x(24)"			/* moduleName restated as its components (calculated)	*/
  field ipName      as character format "x(16)"			/* internal procedure or method name (calculated)	*/
  index idx-moduleId is primary moduleId
  index idx-moduleName moduleName
.

define temp-table tt_lineSummary				/* == line summary section ==				*/
  field moduleId       as integer format ">>>>>9"		/* module id, 0 = session overhead			*/
  field sourceLine     as integer format "->>>>9"		/* line#, 0 = instantiation, -1 = cleanup, -2 = garbage collection (v3+)	*/
  field executionCount as integer format ">>>>>>>>>>9"		/* number of times executed				*/
  field executionTime  as decimal format ">>>9.999999"		/* execution time					*/
  field clockTime      as decimal format ">>>9.999999"		/* cumulative time (clock time)				*/
  /***/
  field averageTime    as decimal format ">>>9.999999"		/* average execution time (calculated)			*/
  index idx-moduleId-sourceLine is unique primary
    moduleId sourceLine
.


procedure loadHeader:

  define variable inFileVersion as integer   no-undo.
  define variable inFileDate    as date      no-undo.
  define variable inFileDesc    as character no-undo.
  define variable inFileTime    as character no-undo.

  repeat:							/* 1 11/07/2020 "Sample Profiling Session" 08:57:37 ""	*/

    inFileVersion = ?.
    import stream inStrm inFileVersion inFileDate inFileDesc inFileTime no-error.

    if  inFileVersion = ? or inFileVersion < 1 then		/* this does not look like a profiler data file...	*/
      do:
        message
          " Unknown profiler version:" inFileVersion skip
          " This does not look like a profiler data file. "
          view-as alert-box information title " Unknown Version "
        .
        keepReading = no.
        return.
      end.

    if  inFileVersion > 4 then					/* this code should be profiler version agnostic	*/
      do:							/* profiler versions are non-breaking			*/
        message							/* so just note that a new version has appeared		*/
          " Unknown profiler version:" inFileVersion skip
          " That is probably not a problem, we are just noting it."
          view-as alert-box information title " Unknown Version "
        .
      end.

    create tt_header.
    assign
      tt_header.profileDate = inFileDate
      tt_header.description = inFileDesc
    .

    keepReading = yes.						/* this file looks like profiler output			*/

  end.

  return.

end.


procedure loadModules:

  if keepReading = no then
    return.

  repeat:
    create tt_module.
    import stream inStrm tt_module.moduleId tt_module.moduleName tt_module.debugName no-error.
    if num-entries( moduleName, " " ) = 1 then
      assign
        srcName = moduleName
        ipName  = ""
      .
     else
      assign
        srcName = entry( 2, moduleName, " " )
        ipName  = entry( 1, moduleName, " " )
      .
  end.

  /* create tt_module. */					/* an extra will be left over from the REPEAT logic	*/

  assign
    tt_module.moduleId   = 0
    tt_module.moduleName = "Session"
    tt_module.debugName  = "Session"
  .

  if debugLevel > 0 then
    do:
      output to value( "modules.txt" ).
      for each tt_module by srcName:
        display tt_module.srcName tt_module.ipName.
      end.
      output close.
    end.

  return.

end.


procedure loadLineSummary:

  if keepReading = no then
    return.

  repeat:							/* 14 4973 71325 0.324374 0.324374			*/
    create tt_lineSummary.
    import stream inStrm tt_lineSummary.moduleId tt_lineSummary.sourceLine tt_lineSummary.executionCount tt_lineSummary.executionTime tt_lineSummary.clockTime no-error.
    tt_lineSummary.averageTime = tt_lineSummary.executionTime / tt_lineSummary.executionCount.
  end.

  delete tt_lineSummary.					/* delete the extra					*/

  if debugLevel > 0 then
    do:
      output to value( "linesummary.txt" ).
      for each tt_lineSummary by tt_lineSummary.moduleId by tt_lineSummary.sourceLine by tt_lineSummary.averageTime descending:     
        find tt_module where tt_module.moduleId = tt_lineSummary.moduleId no-error.
        display
          tt_module.srcName
          tt_lineSummary.sourceLine
          tt_lineSummary.averageTime
        .
      end.
      output close.
    end.

  return.

end.


procedure loadNothing:						/* consume uninteresting sections of the input file	*/

  define variable nothing as character no-undo extent 16.

  if keepReading = no then
    return.

  repeat:
    import stream inStrm nothing no-error.
  end.

  return.

end.


define variable reportColumn as integer no-undo.		/* Expressions in frames cannot use local program variables. (3396)	*/

procedure reportTopX:

  define input parameter outputFileName as character no-undo.
  define input parameter toTTY          as logical   no-undo.
  define input parameter topLines       as integer   no-undo.

  define variable linesShown  as integer no-undo.
  define variable screenWidth as integer no-undo.
  define variable totalXTime  as decimal no-undo.

  screenWidth = 80.

  if session:batch = no and toTTY = yes then
    screenWidth = current-window:width.

  reportColumn = max( 1, (( screenWidth / 2 ) - 40 )).		/* the report itself is 80 characters wide, thus the 40	*/

  for each tt_lineSummary:
    totalXTime = totalXTime + tt_lineSummary.executionTime.
  end.

  if toTTY = no then output to value( outputFileName ).

  display
    tt_header.description format "x(78)" no-label                  skip
    inputFileName         format "x(70)" label    "Input"  colon 8 skip
    outputFileName        format "x(70)" label    "Output" colon 8 skip
    skip(1)
    "Line 0 = procedure initialization, -1 = cleanup, -2 = garbage collection" skip
    skip(1)
   with frame profilerSessionHeader
    no-box
    column reportColumn
    overlay
    side-labels
    stream-io
  .

  linesShown = 0.

  for each tt_lineSummary by tt_lineSummary.executionTime descending:     

    find tt_module where tt_module.moduleId = tt_lineSummary.moduleId no-error.

    linesShown = linesShown + 1.

    display
      ( if available tt_module then tt_module.srcName else string( tt_lineSummary.moduleId )) format "x(24)" label "Module"
      tt_lineSummary.sourceLine     label "Line#"
      tt_lineSummary.executionTime  label "Time"
      tt_lineSummary.averageTime    label "Avg Time"
      (( tt_lineSummary.executionTime / totalXTime ) * 100 ) format ">>9.99%" label "Exec%"
      ( if available tt_module then tt_module.ipName else "**missing**" ) format "x(16)" label "IP/Method"
     with frame profilerSessionReport
      no-box
      column reportColumn
      overlay
      down
      stream-io
    .

    if linesShown > topLines then leave.

  end.

  if toTTY = no then
    output close.
   else
    do:
      pause.
      hide frame profilerSessionHeader.
      hide frame profilerSessionReport.
    end.

  return.

end.


/* main body
 *
 */


if session:window-system <> "tty" then				/* try to look reasonable if someone fires up prowin	*/
  assign
    current-window:hidden       = true
    session:v6display           = true
    current-window:height-chars = 30
    current-window:width-chars  = 90				/* leaves a bit of padding on the sides			*/
    current-window:bgcolor      = 15				/* white background matches the displayed frames	*/
    current-window:hidden       = false
    current-window:visible      = true
  .

if session:parameter = "" or program-name( 4 ) <> ? then 
  do:
    update
       skip(1)
       " " inputFileName format "x(255)" view-as fill-in size 64 by 1  "... " skip
       skip(1)
     with frame getInputFile
      title " Profiler Data Source "
      row 3
      centered
      overlay
      side-labels
    .
    hide frame getInputFile.
  end.
 else
  do:
    if num-entries( session:parameter, "|" ) >= 1 then inputFileName = entry( 1, session:parameter, "|" ).
    if num-entries( session:parameter, "|" ) >= 2 then outputFileName = entry( 2, session:parameter, "|" ).
    if num-entries( session:parameter, "|" ) >= 3 then topLines = integer( entry( 3, session:parameter, "|" )) no-error.
    if num-entries( session:parameter, "|" ) >= 4 then debugLevel = integer( entry( 4, session:parameter, "|" )) no-error.
  end.

file-info:file-name = inputFileName.

if file-info:full-pathname = ? then
  do:

    message
      "Cannot find profiler data file:" inputFileName
      view-as alert-box information title " Cannot Find Input "
    .

    if program-name( 4 ) <> ? then				/* was this procedure started via -p?			*/
      return.							/* nope, someone called us so return there		*/
     else
      quit.							/* yup, do not drop out to the "tram lines"		*/

  end.

if outputFileName = "" then
  do:
    outputFileName = inputFileName.
    if num-entries( outputFileName, "." ) > 1 then
      outputFileName = substring( outputFileName, 1, r-index( outputFileName, "." ) - 1 ).
     outputFileName = outputFileName + ".rpt".
  end.

if topLines = 0 or topLines = ? then
  topLines = 20.

assign
  dateFormat = session:date-format				/* remember the date format				*/
  numberSep  = session:numeric-format				/* numbers too...					*/
  session:date-format    = "mdy"				/* profiler input is always in mdy format		*/
  session:numeric-format = "american"				/* a lazy american must have written this code...	*/
.

input stream inStrm from value( file-info:full-pathname ).

run loadHeader.
run loadModules.
run loadNothing.						/* the call tree is not relevant to our current purpose	*/
run loadLineSummary.

input stream inStrm close.

assign
  session:date-format    = dateFormat				/* restore the original settings			*/
  session:numeric-format = numberSep
.

find first tt_header no-error.
if not available tt_header then
  do:

    message
      "No profiling data was found to analyze"
      view-as alert-box information title " No Data "
    .

  end.
 else
  do:

    run reportTopX( outputFileName, no, topLines ).		/* report on the top X execution time lines -- to file	*/

    if session:batch = no then
      run reportTopX( outputFileName, yes, topLines ).		/* report on the top X execution time lines -- to TTY	*/

  end.

if program-name( 4 ) <> ? then					/* was this procedure started via -p?			*/
  return.							/* nope, someone called us so return there		*/
 else
  quit.								/* yup, do not drop out to the "tram lines"		*/
